use std::collections::HashMap;

use serde::{Deserialize, Serialize};
use thiserror::Error;

use crate::path::{KeyPath, KnobPath};
use crate::style::{KeyStyle, KnobStyle};

#[derive(Debug, Eq, PartialEq, Copy, Clone, Serialize, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum RotationDirection {
    Clockwise,
    Counterclockwise,
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
#[serde(tag = "type", rename_all = "kebab-case")]
pub enum KnobEvent {
    Press,
    ButtonDown,
    ButtonUp,
    Rotate { direction: RotationDirection },
    VisibilityChange { is_visible: bool },
}

#[derive(Debug, PartialEq, Clone, Copy, Serialize, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum KeyTouchEventKind {
    Start,
    Move,
    End,
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
#[serde(tag = "type", rename_all = "kebab-case")]
pub enum KeyEvent {
    Press,
    Touch { touch_id: u8, x: u16, y: u16, kind: KeyTouchEventKind },
    VisibilityChange { is_visible: bool },
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
#[serde(tag = "type", rename_all = "kebab-case")]
pub enum HandlerEvent {
    Stop,
    Knob { path: KnobPath, event: KnobEvent },
    Key { path: KeyPath, event: KeyEvent },
}

#[derive(Debug, PartialEq, Clone, Copy, Serialize, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum VibrationPattern {
    Short,
    Medium,
    Long,
    Low,
    ShortLow,
    ShortLower,
    Lower,
    Lowest,
    DescendSlow,
    DescendMed,
    DescendFast,
    AscendSlow,
    AscendMed,
    AscendFast,
    RevSlowest,
    RevSlow,
    RevMed,
    RevFast,
    RevFaster,
    RevFastest,
    RiseFall,
    Buzz,
    Rumble5,
    Rumble4,
    Rumble3,
    Rumble2,
    Rumble1,
    VeryLong,
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
#[serde(tag = "command", rename_all = "kebab-case")]
pub enum HandlerCommand {
    Vibrate { pattern: VibrationPattern },
    SetKeyStyle { path: KeyPath, value: Option<KeyStyle> },
    SetKnobStyle { path: KnobPath, value: Option<KnobStyle> },
    SetKnobValue { path: KnobPath, value: Option<f32> },
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
pub struct InitialHandlerMessage<GlobalConfig: Clone, KeyConfig: Clone, KnobConfig: Clone> {
    pub global_config: GlobalConfig,
    pub key_configs: HashMap<KeyPath, KeyConfig>,
    pub knob_configs: HashMap<KnobPath, KnobConfig>,
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
#[serde(tag = "type")]
pub enum HandlerInitializationResultMessage {
    Ready,
    Error { error: HandlerInitializationError },
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize, Error)]
pub enum HandlerInitializationError {
    #[error("The provided handler config is invalid: {message}")]
    InvalidConfig {
        requires_global_config: bool,
        supports_keys: bool,
        supports_knobs: bool,
        message: Box<str>,
    },
    #[error("{message}")]
    Other { message: Box<str> },
}
