use deckster_mode::shared::handler_communication::{HandlerEvent, HandlerInitializationError, InitialHandlerMessage, KeyEvent};
use deckster_mode::shared::path::KeyPath;
use deckster_mode::DecksterHandler;
use serde::{Deserialize, Serialize};
use std::fmt::Debug;
use std::thread;
use tokio::process::Command;
use tokio::sync::broadcast;
use tokio::task::LocalSet;

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
pub struct KeyConfig {
    program: Box<str>,
    #[serde(default)]
    args: Box<[String]>,
}

pub struct Handler {
    events_sender: broadcast::Sender<(KeyPath, KeyEvent)>,
}

impl Handler {
    pub fn new(data: InitialHandlerMessage<(), KeyConfig, ()>) -> Result<Self, HandlerInitializationError> {
        let (events_sender, _) = broadcast::channel::<(KeyPath, KeyEvent)>(5);

        thread::spawn({
            let events_sender = events_sender.clone();

            move || {
                let runtime = tokio::runtime::Builder::new_current_thread().enable_io().build().unwrap();
                let task_set = LocalSet::new();

                for (path, config) in data.key_configs {
                    task_set.spawn_local(manage_key(events_sender.subscribe(), path, config));
                }

                runtime.block_on(task_set)
            }
        });

        Ok(Handler { events_sender })
    }
}

impl DecksterHandler for Handler {
    fn handle(&mut self, event: HandlerEvent) {
        if let HandlerEvent::Key { path, event } = event {
            self.events_sender.send((path, event)).unwrap();
        }
    }
}

pub async fn manage_key(mut events: broadcast::Receiver<(KeyPath, KeyEvent)>, path: KeyPath, config: KeyConfig) {
    let mut command = Command::new(config.program.as_ref());
    let command = command.args(config.args.clone().iter());

    while let Ok((p, event)) = events.recv().await {
        if p != path {
            continue;
        }

        if event == KeyEvent::Press {
            tokio::spawn({
                let future = command.status();
                let path = path.clone();

                async move {
                    let result = future.await;

                    match result {
                        Ok(status) => {
                            if !status.success() {
                                log::error!("Process failed with status (key at {}): {}", &path, status)
                            }
                        }
                        Err(error) => log::error!("Failed to spawn process (key at {}): {}", &path, error),
                    }
                }
            });
        }
    }
}
