use futures_util::SinkExt;
use parse_display::{Display, FromStr};
use reqwest::header::{HeaderMap, HeaderValue};
use serde::{Deserialize, Serialize};
use serde_json::json;
use serde_with::{DeserializeFromStr, SerializeDisplay};
use std::cmp::min;
use std::collections::HashMap;
use std::sync::Arc;
use std::time::Duration;
use tokio::sync::{broadcast, RwLock};
use tokio_stream::StreamExt;
use tokio_tungstenite::tungstenite;
use url::Url;

#[derive(Debug, Clone, FromStr, Display, SerializeDisplay, DeserializeFromStr, Eq, PartialEq, Hash)]
#[display("{domain}.{object_id}")]
pub struct EntityId {
    pub domain: String,
    pub object_id: String,
}

#[derive(Debug, Clone)]
pub enum StateUpdate {
    Disconnected,
    Actual(Arc<ActualStateUpdate>),
}

#[derive(Debug)]
pub struct ActualStateUpdate {
    pub entity_id: EntityId,
    pub state: Box<str>,
    pub timestamp: Box<str>,
    pub attributes: serde_json::Value,
}

#[derive(Debug, Clone)]
pub struct HaClient {
    state_updates_sender: broadcast::Sender<StateUpdate>,
    http_client: reqwest::Client,
    base_url: Url,
}

impl HaClient {
    pub async fn new(base_url: Url, token: Box<str>, subscribed_entity_ids: Vec<EntityId>) -> Self {
        let http_client = reqwest::ClientBuilder::new()
            .connect_timeout(Duration::from_secs(10))
            .default_headers({
                let mut map = HeaderMap::new();
                map.insert(
                    "Authorization",
                    HeaderValue::from_str(&format!("Bearer {token}")).expect("the token generated by Home Assistant only contains valid characters"),
                );
                map
            })
            .user_agent(format!("home_assistant deckster handler (v{})", env!("CARGO_PKG_VERSION")))
            .build()
            .unwrap(); // The HTTP client being available is essential.

        let state_updates_sender = broadcast::Sender::<StateUpdate>::new(min(subscribed_entity_ids.len(), 16));
        let state_timestamp_by_entity_id = subscribed_entity_ids.iter().map(|i| (i.clone(), "".to_owned().into_boxed_str())).collect();

        tokio::spawn(do_work(
            base_url.clone(),
            token,
            state_updates_sender.clone(),
            http_client.clone(),
            state_timestamp_by_entity_id,
        ));

        if log::max_level() <= log::Level::Debug {
            let mut updates = state_updates_sender.subscribe();
            tokio::spawn(async move {
                while let Ok(u) = updates.recv().await {
                    log::debug!("State update: {u:?}")
                }
            });
        }

        HaClient {
            state_updates_sender,
            http_client,
            base_url,
        }
    }

    pub fn subscribe_to_state_updates(&self) -> broadcast::Receiver<StateUpdate> {
        self.state_updates_sender.subscribe()
    }

    pub async fn call_service(&self, domain: &str, name: &str, data: impl Into<reqwest::Body>) {
        let result = self
            .http_client
            .post(self.base_url.join(&format!("/api/services/{domain}/{name}")).unwrap())
            .body(data)
            .send()
            .await
            .and_then(|a| a.error_for_status());

        if let Err(error) = result {
            log::error!(
                "POST request to {} failed: {error}",
                error.url().map(|u| u.to_string()).unwrap_or("?".to_owned())
            )
        }
    }

    pub async fn toggle_entity(&self, entity_id: &EntityId) {
        self.call_service(&entity_id.domain, "toggle", json!({ "entity_id": entity_id.to_string() }).to_string())
            .await
    }

    pub async fn press_button_entity(&self, entity_id: &EntityId) {
        self.call_service(&entity_id.domain, "press", json!({ "entity_id": entity_id.to_string() }).to_string())
            .await
    }
}

async fn do_work(
    base_url: Url,
    token: Box<str>,
    state_updates_sender: broadcast::Sender<StateUpdate>,
    http_client: reqwest::Client,
    state_timestamp_by_entity_id: HashMap<EntityId, Box<str>>,
) {
    let states_url = base_url.join("/api/states/").unwrap();
    let websocket_url = {
        let mut u = base_url.clone();
        u.set_scheme(&u.scheme().replace("http", "ws")).unwrap();
        u.set_path("api/websocket");
        u.to_string()
    };

    let mut is_first_connection_attempt = true;
    let state_timestamp_by_entity_id = Arc::new(RwLock::new(state_timestamp_by_entity_id));

    loop {
        let connection_result = tokio_tungstenite::connect_async(&websocket_url).await;

        match connection_result {
            Err(tungstenite::Error::Io(error)) => {
                if is_first_connection_attempt {
                    log::warn!("Establishing a WebSocket connection failed: {error}");
                    log::info!("Retrying every 5 seconds…")
                }

                is_first_connection_attempt = false;
                tokio::time::sleep(Duration::from_secs(5)).await;
            }
            Err(error) => panic!("WebSocket error: {}", error),
            Ok((mut socket, _)) => {
                log::info!("WebSocket connection successfully established.");

                while let Some(event) = socket.next().await {
                    match event {
                        Err(error) => {
                            log::error!("The WebSocket connection failed: {error}");
                            state_updates_sender.send(StateUpdate::Disconnected).unwrap();

                            break;
                        }
                        Ok(message) => match message {
                            tungstenite::Message::Ping(data) => socket.send(tungstenite::Message::Pong(data)).await.unwrap(),
                            tungstenite::Message::Text(data) => {
                                log::trace!("Received WebSocket message: {data}");

                                match serde_json::from_str::<HaIncomingWsMessage>(&data) {
                                    Err(error) => log::error!("Deserializing WebSocket message failed: {error}"),
                                    Ok(message) => match message {
                                        HaIncomingWsMessage::AuthRequired { .. } => socket
                                            .send(tungstenite::Message::Text(
                                                serde_json::to_string(&HaOutgoingWsMessage::Auth { access_token: token.clone() }).unwrap(),
                                            ))
                                            .await
                                            .unwrap(),
                                        HaIncomingWsMessage::AuthInvalid { .. } => panic!("Invalid access token."),
                                        HaIncomingWsMessage::AuthOk { .. } => {
                                            let subscription_message = serde_json::to_string_pretty(&HaOutgoingWsMessage::SubscribeTrigger {
                                                // ID may not be zero (that one took me a while)
                                                id: 1,
                                                trigger: HaTrigger::State {
                                                    entity_id: state_timestamp_by_entity_id.read().await.keys().cloned().collect(),
                                                },
                                            })
                                            .unwrap();

                                            socket.send(tungstenite::Message::Text(subscription_message)).await.unwrap();
                                        }
                                        HaIncomingWsMessage::Result { id, success } => {
                                            if !success {
                                                panic!("A command ({}) failed.", id);
                                            }

                                            for entity_id in state_timestamp_by_entity_id.read().await.keys() {
                                                tokio::spawn(request_entity_state(
                                                    states_url.join(&entity_id.to_string()).unwrap(),
                                                    http_client.clone(),
                                                    Arc::clone(&state_timestamp_by_entity_id),
                                                    state_updates_sender.clone(),
                                                ));
                                            }
                                        }
                                        HaIncomingWsMessage::Event { event, .. } => match extract_state_update_from_event(&event) {
                                            None => log::error!("Invalid state change event message: {data}"),
                                            Some(update) => {
                                                // LOCK START
                                                let mut state_timestamp_by_entity_id = state_timestamp_by_entity_id.write().await;

                                                match state_timestamp_by_entity_id.get(&update.entity_id) {
                                                    None => log::warn!("Received unwanted state change event for entity '{}'", update.entity_id),
                                                    Some(last_timestamp) => {
                                                        if last_timestamp < &update.timestamp {
                                                            state_timestamp_by_entity_id.insert(update.entity_id.clone(), update.timestamp.clone());
                                                            state_updates_sender.send(StateUpdate::Actual(Arc::new(update))).unwrap();
                                                        }
                                                    }
                                                }
                                                // LOCK END
                                            }
                                        },
                                    },
                                }
                            }
                            _ => log::error!("Received unsupported WebSocket message: {message:?}"),
                        },
                    }
                }
            }
        };
    }
}

async fn request_entity_state(
    url: Url,
    http_client: reqwest::Client,
    state_timestamp_by_entity_id: Arc<RwLock<HashMap<EntityId, Box<str>>>>,
    state_updates_sender: broadcast::Sender<StateUpdate>,
) {
    match http_client.get(url).send().await.and_then(|a| a.error_for_status()) {
        Err(error) => log::error!(
            "A GET request to {} failed: {error}",
            error.url().map(|u| u.to_string()).unwrap_or("?".to_owned())
        ),
        Ok(response) => match serde_json::from_str(&response.text().await.unwrap()) {
            Ok(object) => match extract_state_update_from_state(&object) {
                None => log::error!("Invalid entity state object: {object}"),
                Some(update) => {
                    // LOCK START
                    let mut state_timestamp_by_entity_id = state_timestamp_by_entity_id.write().await;
                    let last_timestamp = state_timestamp_by_entity_id
                        .get(&update.entity_id)
                        .expect("Home Assistant responds with the state of the requested entity.");

                    if last_timestamp < &update.timestamp {
                        state_timestamp_by_entity_id.insert(update.entity_id.clone(), update.timestamp.clone());
                        state_updates_sender.send(StateUpdate::Actual(Arc::new(update))).unwrap();
                    }
                    // LOCK END
                }
            },
            Err(error) => {
                log::error!("Failed to deserialize state object: {error}");
            }
        },
    }
}

fn extract_state_update_from_event(object: &serde_json::Value) -> Option<ActualStateUpdate> {
    extract_state_update_from_state(object.get("variables")?.get("trigger")?.get("to_state")?)
}

fn extract_state_update_from_state(object: &serde_json::Value) -> Option<ActualStateUpdate> {
    Some(ActualStateUpdate {
        state: object.get("state")?.as_str()?.to_owned().into_boxed_str(),
        entity_id: object.get("entity_id")?.as_str()?.parse().ok()?,
        timestamp: object.get("last_updated")?.as_str()?.to_owned().into_boxed_str(),
        attributes: object.get("attributes")?.to_owned(),
    })
}

#[derive(Debug, Deserialize)]
#[serde(tag = "type", rename_all = "snake_case")]
pub enum HaIncomingWsMessage {
    AuthRequired { ha_version: Box<str> },
    AuthOk { ha_version: Box<str> },
    AuthInvalid { message: Box<str> },
    Result { id: usize, success: bool },
    Event { id: usize, event: serde_json::Value },
}

#[derive(Debug, Serialize)]
#[serde(tag = "type", rename_all = "snake_case")]
pub enum HaOutgoingWsMessage {
    Auth { access_token: Box<str> },
    SubscribeTrigger { id: usize, trigger: HaTrigger },
}

#[derive(Debug, Serialize)]
#[serde(tag = "platform", rename_all = "snake_case")]
pub enum HaTrigger {
    State { entity_id: Box<[EntityId]> },
}
