use std::collections::HashMap;
use std::ffi::OsString;
use std::path::Path;
use std::process::{exit, Stdio};
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::Arc;

use color_eyre::eyre::{eyre, WrapErr};
use color_eyre::Result;
use is_executable::IsExecutable;
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::process::{ChildStdin, Command};

use deckster_shared::handler_communication::{
    HandlerCommand, HandlerEvent, HandlerInitializationError, HandlerInitializationResultMessage, InitialHandlerMessage,
};
use deckster_shared::path::{KeyPath, KnobPath};

use crate::model::mqtt::HandlerHostsConfig;

pub async fn start(
    host_id: Box<str>,
    handlers_directory: &Path,
    global_handler_configs: &HashMap<Box<str>, toml::Table>,
    handler_hosts_config: HandlerHostsConfig,
    commands_sender: flume::Sender<HandlerCommand>,
    mut events_receiver: tokio::sync::broadcast::Receiver<HandlerEvent>,
) -> Result<()> {
    let should_stop = Arc::new(AtomicBool::new(false));

    let handler_names: Vec<OsString> = std::fs::read_dir(handlers_directory)
        .wrap_err_with(|| format!("while reading the handlers directory: {}", handlers_directory.to_string_lossy()))?
        .filter_map(|entry| {
            if let Ok(entry) = entry {
                let path = entry.path();
                if path.is_executable() {
                    return Some(path.file_name().expect("the path does not end in `..`").to_os_string());
                }
            }

            None
        })
        .collect();

    let mut handler_stdin_by_name: HashMap<Box<str>, ChildStdin> = HashMap::with_capacity(handler_names.len());

    let mut handler_name_by_key_path: HashMap<KeyPath, Box<str>> = HashMap::new();
    let mut handler_config_by_key_path_by_handler_name: HashMap<Box<str>, HashMap<KeyPath, Arc<toml::Table>>> = HashMap::new();

    for (path, config) in handler_hosts_config.keys {
        if config.host_id != host_id {
            continue;
        }

        handler_name_by_key_path.insert(path.clone(), config.name.clone());
        handler_config_by_key_path_by_handler_name
            .entry(config.name)
            .or_default()
            .insert(path, Arc::clone(&config.config));
    }

    let mut handler_name_by_knob_path: HashMap<KnobPath, Box<str>> = HashMap::new();
    let mut handler_config_by_knob_path_by_handler_name: HashMap<Box<str>, HashMap<KnobPath, Arc<toml::Table>>> = HashMap::new();

    for (path, config) in handler_hosts_config.knobs {
        if config.host_id != host_id {
            continue;
        }

        handler_name_by_knob_path.insert(path.clone(), config.name.clone());
        handler_config_by_knob_path_by_handler_name
            .entry(config.name)
            .or_default()
            .insert(path, Arc::clone(&config.config));
    }

    for handler_name in handler_names {
        let handler_name = handler_name
            .into_string()
            .map_err(|string| {
                eyre!(
                    "Handler names must be valid Unicode. Offending name (with non-Unicode characters replaced): {}",
                    string.to_string_lossy()
                )
            })?
            .into_boxed_str();

        start_handler(
            handlers_directory,
            global_handler_configs.get(&handler_name),
            &mut handler_config_by_key_path_by_handler_name,
            &mut handler_config_by_knob_path_by_handler_name,
            &mut handler_stdin_by_name,
            commands_sender.clone(),
            Arc::clone(&should_stop),
            handler_name.clone(),
        )
        .await
        .wrap_err_with(|| format!("while starting handler: {handler_name}"))?;
    }

    if let Some((handler_name, config_by_key_path)) = handler_config_by_key_path_by_handler_name.drain().next() {
        return Err(eyre!(
            "There is no executable file named '{handler_name}' in the handlers directory but these keys have it set as their handler: {}",
            config_by_key_path.keys().join(", ")
        ));
    }

    if let Some((handler_name, config_by_knob_path)) = handler_config_by_knob_path_by_handler_name.drain().next() {
        return Err(eyre!(
            "There is no executable file named '{handler_name}' in the handlers directory but these knobs have it set as their handler: {}",
            config_by_knob_path.keys().join(", ")
        ));
    }

    tokio::spawn(async move {
        while let Ok(event) = events_receiver.recv().await {
            let result = forward_event_to_handler(
                &should_stop,
                &mut handler_stdin_by_name,
                &handler_name_by_key_path,
                &handler_name_by_knob_path,
                event,
            )
            .await;

            match result {
                Err(error) => {
                    log::error!("Failed to forward an event to its handler: {error}")
                }
                Ok(should_continue) => {
                    if !should_continue {
                        break;
                    }
                }
            }
        }
    });

    Ok(())
}

async fn start_handler(
    handlers_directory: &Path,
    global_config: Option<&toml::Table>,
    handler_config_by_key_path_by_handler_name: &mut HashMap<Box<str>, HashMap<KeyPath, Arc<toml::Table>>>,
    handler_config_by_knob_path_by_handler_name: &mut HashMap<Box<str>, HashMap<KnobPath, Arc<toml::Table>>>,
    handler_stdin_by_name: &mut HashMap<Box<str>, ChildStdin>,
    commands_sender: flume::Sender<HandlerCommand>,
    should_stop: Arc<AtomicBool>,
    handler_name: Box<str>,
) -> Result<()> {
    let (key_configs, knob_configs) = match (
        handler_config_by_key_path_by_handler_name.remove(&handler_name),
        handler_config_by_knob_path_by_handler_name.remove(&handler_name),
    ) {
        (None, None) => {
            log::warn!("Handler '{handler_name}' is not used by any key or knob.");
            return Ok(());
        }
        (a, b) => (a.unwrap_or_default(), b.unwrap_or_default()),
    };

    let mut command = Command::new(handlers_directory.join(handler_name.to_string()))
        .arg("deckster-run")
        .stdin(Stdio::piped())
        .stdout(Stdio::piped())
        .spawn()
        .wrap_err("while spawning the handler process")?;

    let mut stdout_lines = BufReader::new(command.stdout.take().expect("stdout is explicitly captured and has not yet been taken")).lines();
    let mut stdin = command.stdin.take().expect("stdin is explicitly captured and has not yet been taken");

    let initial_handler_message = InitialHandlerMessage {
        global_config,
        key_configs,
        knob_configs,
    };

    let serialized_message = serde_json::to_string(&initial_handler_message).unwrap().into_boxed_str().into_boxed_bytes();

    stdin.write_all(&serialized_message).await.wrap_err("while writing to stdin")?;
    stdin.write_u8(b'\n').await.wrap_err("while writing to stdin")?;
    stdin.flush().await.wrap_err("while flushing stdin")?;

    // TODO: It is not clear when the returned Option is None. We just unwrap and hope for the best.
    let result_line = stdout_lines.next_line().await.wrap_err("while reading from stdout")?.unwrap();
    let result: HandlerInitializationResultMessage = serde_json::from_str(&result_line)?;

    if let HandlerInitializationResultMessage::Error { error } = result {
        #[rustfmt::skip]
        if let HandlerInitializationError::InvalidConfig { requires_global_config, supports_keys, supports_knobs, .. } = error {
            if requires_global_config && initial_handler_message.global_config.is_none() {
                return Err(eyre!("The '{handler_name}' handler requires a global configuration in the deckster.toml file."));
            } else if !supports_keys && !initial_handler_message.key_configs.is_empty() {
                return Err(eyre!(
                    "The '{handler_name}' handler does not support keys, but these keys tried to use it: {}",
                    initial_handler_message.key_configs.keys().map(|k| k.to_string()).join(", ")
                ));
            } else if !supports_knobs && !initial_handler_message.knob_configs.is_empty() {
                return Err(eyre!(
                    "The '{handler_name}' handler does not support knobs, but these knobs tried to use it: {}",
                    initial_handler_message.knob_configs.keys().map(|k| k.to_string()).join(", ")
                ));
            }
        };

        return Err(eyre!("Starting the '{handler_name}' handler failed: {error}"));
    }

    handler_stdin_by_name.insert(handler_name.clone(), stdin);

    tokio::spawn(async move {
        while let Ok(Some(line)) = stdout_lines.next_line().await {
            if line.starts_with('{') {
                if let Ok(command) = serde_json::from_str::<HandlerCommand>(&line) {
                    commands_sender.send_async(command).await.unwrap();
                } else {
                    log::error!("Failed to deserialize command object: {line}");
                }
            } else {
                println!("{}", line);
            }
        }
    });

    let should_stop = Arc::clone(&should_stop);
    tokio::spawn(async move {
        let exit_status = command.wait().await.unwrap();

        if !should_stop.load(Ordering::Relaxed) {
            match exit_status.code() {
                None => log::error!("The '{handler_name}' handler was unexpectedly terminated by a signal."),
                Some(code) => log::error!("The '{handler_name}' handler exited unexpectedly with status code {code}"),
            }

            exit(1);
        } else {
            log::debug!("The '{handler_name}' handler exited: {exit_status:#?}");
        }
    });

    Ok(())
}

async fn forward_event_to_handler(
    should_stop: &AtomicBool,
    handler_stdin_by_name: &mut HashMap<Box<str>, ChildStdin>,
    handler_name_by_key_path: &HashMap<KeyPath, Box<str>>,
    handler_name_by_knob_path: &HashMap<KnobPath, Box<str>>,
    event: HandlerEvent,
) -> Result<bool> {
    let handler_name = match &event {
        HandlerEvent::Stop => {
            should_stop.store(true, Ordering::Relaxed);
            let serialized_event = serde_json::to_string(&event).unwrap().into_boxed_str().into_boxed_bytes();
            for handler_stdin in handler_stdin_by_name.values_mut() {
                // We assume that stdin being closed means that the handler process is no longer running.
                // If that is the case, we do not need to stop it as it is already stopped.
                _ = handler_stdin.write_all(&serialized_event).await;
                _ = handler_stdin.write_u8(b'\n').await;
                _ = handler_stdin.flush().await;
            }

            return Ok(false);
        }
        HandlerEvent::Key { path, .. } => handler_name_by_key_path.get(path),
        HandlerEvent::Knob { path, .. } => handler_name_by_knob_path.get(path),
    };

    if let Some(handler_name) = handler_name {
        let handler_stdin = handler_stdin_by_name.get_mut(handler_name).expect("was already checked above");
        let serialized_event = serde_json::to_string(&event).unwrap().into_boxed_str().into_boxed_bytes();

        handler_stdin.write_all(&serialized_event).await.wrap_err("while writing to stdin")?;
        handler_stdin.write_u8(b'\n').await.wrap_err("while writing to stdin")?;
        handler_stdin.flush().await.wrap_err("while flushing stdin")?;
    }

    Ok(true)
}

#[derive(Debug, Serialize, Deserialize)]
pub struct KeyOrKnobHandlerConfig {
    pub host_id: Box<str>,
    pub name: Box<str>,
    pub config: Arc<toml::Table>,
}
